from lisa.server.web.manageplugins.models import Plugin, Description, Rule, Cron, Intent
import json, git, os, shutil
from shutil import rmtree
from twisted.python.reflect import namedAny
import inspect
from django.template.loader import render_to_string
import datetime
from lisa.server.web.weblisa.settings import LISA_PATH, configuration

def install(plugin_url=None, plugin_sha=None, plugin_name=None):
    if plugin_url and plugin_sha:
        repo = git.Repo.clone_from(plugin_url, configuration['plugin_path'] + '/' + plugin_name)
        repo.git.checkout(plugin_sha)
    jsonfile = configuration['plugin_path'] + '/' + plugin_name + '/' + str(plugin_name).lower() + '.json'
    metadata = json.load(open(jsonfile))

    if Plugin.objects(name=plugin_name):
        return {'status': 'fail', 'log': 'Plugin already installed'}
    else:
        plugin = Plugin()
        description_list = []
        for item in metadata:
            if item != 'cron' or item != 'rules':
                if item == 'description':
                    for description in metadata[item]:
                        oDescription = Description()
                        for k,v in description.iteritems():
                            setattr(oDescription, k, v)
                        description_list.append(oDescription)
                    setattr(plugin, item, description_list)
                elif item == 'enabled':
                    if metadata[item] == 0:
                        setattr(plugin, item, False)
                    else:
                        setattr(plugin, item, True)
                else:
                    setattr(plugin, item, metadata[item])
        plugin.save()

        for item in metadata:
            if item == 'rules':
                for rule_item in metadata['rules']:
                    rule = Rule()
                    for parameter in rule_item:
                        setattr(rule, parameter, rule_item[parameter])
                    rule.plugin = plugin
                    rule.save()
            if item == 'crons':
                for cron_item in metadata['crons']:
                    cron = Cron()
                    for parameter in cron_item:
                        setattr(cron, parameter, cron_item[parameter])
                    cron.plugin = plugin
                    cron.save()


        for intent, value in metadata['configuration']['intents'].iteritems():
            oIntent = Intent()
            oIntent.name = intent
            oIntent.function = value['method']
            oIntent.module = '.'.join([plugin_name, 'modules', plugin_name.lower(), plugin_name])
            oIntent.enabled = True
            oIntent.plugin = plugin
            oIntent.save()
        os.remove(jsonfile)
        return {'status': 'success', 'log': 'Plugin installed'}


def enable(plugin_name=None, plugin_pk=None):
    if plugin_pk:
        plugin_list = Plugin.objects(pk=plugin_pk)
    else:
        plugin_list = Plugin.objects(name=plugin_name)
    for plugin in plugin_list:
        if plugin.enabled:
            return {'status': 'fail', 'log': 'Plugin already enabled'}
        else:
            plugin.enabled = True
            plugin.save()
            for cron in Cron.objects(plugin=plugin):
                cron.enabled = True
                cron.save()
            for rule in Rule.objects(plugin=plugin):
                rule.enabled = True
                rule.save()

            intent_list = Intent.objects(plugin=plugin)
            for oIntent in intent_list:
                oIntent.enabled = True
                oIntent.save()
            return {'status': 'success', 'log': 'Plugin enabled'}

def disable(plugin_name=None, plugin_pk=None):
    if plugin_pk:
        plugin_list = Plugin.objects(pk=plugin_pk)
    else:
        plugin_list = Plugin.objects(name=plugin_name)
    for plugin in plugin_list:
        if not plugin.enabled:
            return {'status': 'fail', 'log': 'Plugin already disabled'}
        else:
            plugin.enabled = False
            plugin.save()
            for cron in Cron.objects(plugin=plugin):
                cron.enabled = False
                cron.save()
            for rule in Rule.objects(plugin=plugin):
                rule.enabled = False
                rule.save()

            intent_list = Intent.objects(plugin=plugin)
            for oIntent in intent_list:
                oIntent.enabled = False
                oIntent.save()

            return {'status': 'success', 'log': 'Plugin disabled'}


def uninstall(plugin_name=None, plugin_pk=None):
    if plugin_pk:
        plugin_list = Plugin.objects(pk=plugin_pk)
    else:
        plugin_list = Plugin.objects(name=plugin_name)
    if not plugin_list:
        return {'status': 'fail', 'log': 'Plugin not installed'}
    else:
        for plugin in plugin_list:
            rmtree(configuration['plugin_path'] + '/' + plugin['name'])
            plugin.delete()
            for cron in Cron.objects(plugin=plugin):
                cron.delete()
            for rule in Rule.objects(plugin=plugin):
                rule.delete()
            intent_list = Intent.objects(plugin=plugin)
            for oIntent in intent_list:
                oIntent.delete()

        return {'status': 'success', 'log': 'Plugin uninstalled'}


def method_list(plugin_name=None):
    if plugin_name:
        plugin_list = Plugin.objects(name=plugin_name)
    else:
        plugin_list = Plugin.objects.all()
    listallmethods = []
    for plugin in plugin_list:
        plugininstance = namedAny('.'.join((str(plugin.name), 'modules', str(plugin.name).lower(),
                                            str(plugin.name))))()
        listpluginmethods = []
        for m in inspect.getmembers(plugininstance, predicate=inspect.ismethod):
            if not "__init__" in m:
                listpluginmethods.append(m[0])
        listallmethods.append({ 'plugin': plugin.name, 'methods': listpluginmethods})
    for f in os.listdir(os.path.normpath(LISA_PATH + '/core')):
        fileName, fileExtension = os.path.splitext(f)
        if os.path.isfile(os.path.join(os.path.normpath(LISA_PATH + '/core'), f)) and not f.startswith('__init__') and fileExtension != '.pyc':
            coreinstance = namedAny('.'.join(('core', str(fileName).lower(), str(fileName).capitalize())))()
            listcoremethods = []
            for m in inspect.getmembers(coreinstance, predicate=inspect.ismethod):
                #init shouldn't be listed in methods and _ is for translation
                if not "__init__" in m:
                    listcoremethods.append(m[0])
            listallmethods.append({ 'core': fileName, 'methods': listcoremethods})
    print listallmethods
    return listallmethods


def _template_to_file(filename, template, context):
    import codecs
    codecs.open(filename, 'w', 'utf-8').write(render_to_string(template, context))


def create(plugin_name, author_name, author_email):
    import requests
    import pytz

    metareq = requests.get('https://raw.github.com/Seraf/LISA-Plugins/master/plugin_list.json')
    if(metareq.ok):
        for item in json.loads(metareq.text or metareq.content):
            if item['name'] == plugin_name:
                return {'status': 'fail', 'log': 'Plugin already exist in the store'}
    context = {
        'plugin_name': plugin_name,
        'plugin_name_lower': plugin_name.lower(),
        'author_name': author_name,
        'author_email': author_email,
        'creation_date': pytz.UTC.localize(datetime.datetime.now()).strftime("%Y-%m-%d %H:%M%z")
    }
    os.mkdir(os.path.normpath(configuration['plugin_path'] + '/' + plugin_name))

    # Lang stuff
    os.mkdir(os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/lang'))
    os.mkdir(os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/lang/en'))
    os.mkdir(os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/lang/en/LC_MESSAGES'))
    _template_to_file(filename=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/lang/en/LC_MESSAGES/' +
                                                plugin_name.lower() + '.po'),
                      template='plugin/lang/en/LC_MESSAGES/module.po',
                      context=context)

    # Module stuff
    os.mkdir(os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/modules'))
    _template_to_file(filename=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/modules/' +
                                                plugin_name.lower() + '.py'),
                      template='plugin/modules/module.py',
                      context=context)
    open(os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/modules/__init__.py'), "a")

    # Web stuff
    os.mkdir(os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/web'))
    os.mkdir(os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/web/templates'))
    shutil.copy(src=os.path.normpath(LISA_PATH + '/web/manageplugins/templates/plugin/web/templates/widget.html'),
                dst=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/web/templates/widget.html'))
    shutil.copy(src=os.path.normpath(LISA_PATH + '/web/manageplugins/templates/plugin/web/templates/index.html'),
                dst=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/web/templates/index.html'))
    open(os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/web/__init__.py'), "a")
    _template_to_file(filename=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/web/api.py'),
                      template='plugin/web/api.py',
                      context=context)
    _template_to_file(filename=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/web/models.py'),
                      template='plugin/web/models.py',
                      context=context)
    _template_to_file(filename=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/web/tests.py'),
                          template='plugin/web/tests.py',
                          context=context)
    _template_to_file(filename=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/web/urls.py'),
                          template='plugin/web/urls.py',
                          context=context)
    _template_to_file(filename=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/web/views.py'),
                      template='plugin/web/views.py',
                      context=context)

    # Plugin stuff (metadata)
    _template_to_file(filename=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/__init__.py'),
                      template='plugin/__init__.py',
                      context=context)
    _template_to_file(filename=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/README.rst'),
                      template='plugin/README.rst',
                      context=context)
    _template_to_file(filename=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name + '/.gitignore'),
                      template='plugin/.gitignore',
                      context=context)
    _template_to_file(filename=os.path.normpath(configuration['plugin_path'] + '/' + plugin_name +
                                                '/' + plugin_name.lower() + '.json'),
                      template='plugin/module.json',
                      context=context)

    return {'status': 'success', 'log': 'Plugin created'}



from twisted.python import log

#try:
from lisa.server.libs.wit import Wit
from lisa.server.libs.rulesengine import RulesEngine
from lisa.server.libs.commands import Commands
from lisa.server.libs.websocket import LisaClientFactory, WebSocketProtocol
from lisa.server.libs.webserver import verifyCallback, Root
from lisa.server.libs.server import Lisa, LisaFactory, ServerTLSContext, LisaInstance, taskman, scheduler, LisaProtocolInstance, configuration, Initialize
#except ImportError:
#    log.err(ImportError)


# -*- coding: UTF-8 -*-
import os
import json
import sys
from twisted.internet.protocol import Factory, Protocol
from twisted.internet import reactor, ssl
from twisted.application import internet, service
from twisted.web import server, wsgi, static
from twisted.python import threadpool, log
from autobahn.twisted.websocket import WebSocketServerProtocol, WebSocketServerFactory
from autobahn.twisted.resource import WebSocketResource
from OpenSSL import SSL

import pkg_resources
configuration = json.load(open(pkg_resources.resource_filename(__name__, 'configuration/lisa.json.sample')))
path = os.path.abspath(__file__)
dir_path = os.path.dirname(path)

class ThreadPoolService(service.Service):
    def __init__(self, pool):
        self.pool = pool

    def startService(self):
        service.Service.startService(self)
        self.pool.start()

    def stopService(self):
        service.Service.stopService(self)
        self.pool.stop()


# Twisted Application Framework setup:
application = service.Application('LISA')

def makeService(config):
    from django.core.handlers.wsgi import WSGIHandler
    os.environ['DJANGO_SETTINGS_MODULE'] = 'lisa.server.web.weblisa.settings'

    global configuration

    if config['configuration']:
        configuration = json.load(open(config['configuration']))

    # Check if plugin directory exists. If not, create it
    try:
        if not os.path.exists(os.path.normpath(config['plugins'])):
            os.makedirs(os.path.normpath(config['plugins']))
        if not os.path.exists(os.path.normpath(config['plugins'] + '/__init__.py')):
            open(os.path.normpath(config['plugins'] + '/__init__.py'), 'w').close()
    except:
        log.err("Directory %s doesn't exist, and it seems impossible to create it" % config['plugins'])
        pass
    sys.path.append(os.path.normpath(config['plugins']))
    configuration['plugins'] = config['plugins']

    from lisa.server import libs

    # Creating MultiService
    multi = service.MultiService()
    pool = threadpool.ThreadPool()
    tps = ThreadPoolService(pool)
    tps.setServiceParent(multi)

    # Creating the web stuff
    resource_wsgi = wsgi.WSGIResource(reactor, tps.pool, WSGIHandler())
    root = libs.Root(resource_wsgi)
    staticrsrc = static.File('/'.join([dir_path,'web/interface/static']))
    root.putChild("static", staticrsrc)

    # Create the websocket
    if configuration['enable_secure_mode']:
        socketfactory = WebSocketServerFactory("wss://" + configuration['lisa_url'] + ":" +
                                               str(configuration['lisa_web_port_ssl']),debug=False)
    else:
        socketfactory = WebSocketServerFactory("ws://" + configuration['lisa_url'] + ":" +
                                               str(configuration['lisa_web_port']),debug=False)
    socketfactory.protocol = libs.WebSocketProtocol
    socketfactory.protocol.configuration, socketfactory.protocol.dir_path = configuration, dir_path
    socketresource = WebSocketResource(socketfactory)
    root.putChild("websocket", socketresource)

    # Configuring servers to launch
    if configuration['enable_secure_mode'] or configuration['enable_unsecure_mode']:
        if configuration['enable_secure_mode']:
            SSLContextFactoryEngine = ssl.DefaultOpenSSLContextFactory(
                os.path.normpath(dir_path + '/' + 'configuration/ssl/server.key'),
                os.path.normpath(dir_path + '/' + 'configuration/ssl/server.crt')
            )
            SSLContextFactoryWeb = ssl.DefaultOpenSSLContextFactory(
                os.path.normpath(dir_path + '/' + 'configuration/ssl/server.key'),
                os.path.normpath(dir_path + '/' + 'configuration/ssl/server.crt')
            )
            ctx = SSLContextFactoryEngine.getContext()
            ctx.set_verify(
                SSL.VERIFY_PEER | SSL.VERIFY_FAIL_IF_NO_PEER_CERT,
                libs.verifyCallback
            )
            # Since we have self-signed certs we have to explicitly
            # tell the server to trust them.
            with open(os.path.normpath(dir_path + '/' + 'configuration/ssl/server.pem'), 'w') as outfile:
                for file in os.listdir(os.path.normpath(dir_path + '/' + 'configuration/ssl/public/')):
                    with open(os.path.normpath(dir_path + '/' + 'configuration/ssl/public/'+file)) as infile:
                        for line in infile:
                            outfile.write(line)


            ctx.load_verify_locations(os.path.normpath(dir_path + '/' + 'configuration/ssl/server.pem'))
            internet.SSLServer(configuration['lisa_web_port_ssl'],
                               server.Site(root), SSLContextFactoryWeb).setServiceParent(multi)
            internet.SSLServer(configuration['lisa_engine_port_ssl'],
                               libs.LisaInstance, SSLContextFactoryEngine).setServiceParent(multi)
        if configuration['enable_unsecure_mode']:
            # Serve it up:
            internet.TCPServer(configuration['lisa_web_port'], server.Site(root)).setServiceParent(multi)
            internet.TCPServer(configuration['lisa_engine_port'], libs.LisaInstance).setServiceParent(multi)

    else:
        exit(1)
    libs.scheduler.setServiceParent(multi)
    multi.setServiceParent(application)
    libs.Initialize()
    return multi



# -*- coding: UTF-8 -*-
import os, inspect
from pymongo import MongoClient
from lisa.server.service import configuration

import gettext

path = os.path.realpath(os.path.abspath(os.path.join(os.path.split(
    inspect.getfile(inspect.currentframe()))[0],os.path.normpath("../lang/"))))
_ = translation = gettext.translation(domain='{{ plugin_name_lower }}', localedir=path, languages=[configuration['lang']]).ugettext

class {{ plugin_name }}:
    def __init__(self, lisa=None):
        self.lisa = lisa
        self.configuration_lisa = configuration
        self.mongo = MongoClient(self.configuration_lisa['database']['server'],
                            self.configuration_lisa['database']['port'])
        self.configuration = self.mongo.lisa.plugins.find_one({"name": "{{ plugin_name }}"})
        self.build_default_list()
        self.answer = None
        self.raw = None




# -*- coding: UTF-8 -*-
import os, json, sys, uuid
from lisa.server.libs import RulesEngine, Commands, Wit
from twisted.internet.protocol import Factory, Protocol
from twisted.protocols.basic import LineReceiver
from twisted.internet import ssl
from twisted.python import log
from pymongo import MongoClient
from OpenSSL import SSL
from lisa.server.libs.txscheduler.manager import ScheduledTaskManager
from lisa.server.libs.txscheduler.service import ScheduledTaskService

from lisa.server.service import configuration, dir_path
from lisa.server.web.manageplugins.models import Intent, Rule

# Create a task manager to pass it to other services
taskman = ScheduledTaskManager(configuration)
scheduler = ScheduledTaskService(taskman)

enabled_plugins = []

class ServerTLSContext(ssl.DefaultOpenSSLContextFactory):
    def __init__(self, *args, **kw):
        kw['sslmethod'] = SSL.TLSv1_METHOD
        ssl.DefaultOpenSSLContextFactory.__init__(self, *args, **kw)

class Lisa(LineReceiver):
    def __init__(self,factory, wit):
        self.factory = factory
        self.wit = wit

    def answerToClient(self, jsondata):
        if configuration['debug']['debug_output']:
            log.msg("OUTPUT: " + str(jsondata))
        jsonreturned = json.loads(jsondata)
        if 'all' in jsonreturned['clients_zone']:
            for client in self.factory.clients:
                client['object'].sendLine(jsondata)
        else:
            for zone in jsonreturned['clients_zone']:
                if zone == 'sender':
                    self.sendLine(jsondata)
                else:
                    for client in self.factory.clients:
                        if client['zone'] == zone:
                            client['object'].sendLine(jsondata)

    def askToClient(self, jsondata):
        jsondata['type'] = 'question'
        self.answerToClient(jsondata=jsondata)

    def connectionMade(self):
        self.client_uuid = str(uuid.uuid1())
        self.factory.clients.append({"object": self, "zone": "", "type": "", "uuid": self.client_uuid})
        if configuration['enable_secure_mode']:
            ctx = ServerTLSContext(
                privateKeyFileName=os.path.normpath(dir_path + '/' + 'configuration/ssl/server.key'),
                certificateFileName= os.path.normpath(dir_path + '/' + 'configuration/ssl/server.crt')
            )
            self.transport.startTLS(ctx, self.factory)
            pass

    def connectionLost(self, reason):
        log.err('Lost connection.  Reason:', reason)
        for client in self.factory.clients:
            if client['object'] == self:
                self.factory.clients.remove(client)

    def lineReceived(self, data):
        if configuration['debug']['debug_input']:
            log.msg("INPUT: " + str(data))
        try:
            jsonData = json.loads(data)
        except ValueError, e:
            self.sendLine("Error : Invalid JSON")
        else:
            for client in self.factory.clients:
                if client['object'] == self and (not client['type'] or not client['zone']):
                    client['type'], client['zone'] = jsonData['type'], jsonData['zone']
            if jsonData['type'] == "chat":
                RulesEngine(configuration).Rules(jsonData=jsonData, lisaprotocol=self)
            elif jsonData['type'] == "command":
                Commands(configuration, lisaprotocol=self).parse(jsonData=jsonData)

class LisaFactory(Factory):
    def __init__(self):
        self.wit = Wit(configuration)
        self.clients = []
        self.syspath = sys.path
        mongo = MongoClient(configuration['database']['server'], configuration['database']['port'])
        self.database = mongo.lisa
        self.build_activeplugins()

    def build_activeplugins(self):
        global enabled_plugins
        # Load enabled plugins for the main language
        for plugin in self.database.plugins.find( { "enabled": True, "lang": configuration['lang'] } ):
            enabled_plugins.append(str(plugin['name']))

    def buildProtocol(self, addr):
        self.Lisa = Lisa(self,self.wit)
        return self.Lisa

    def LisaReload(self):
        global enabled_plugins
        log.msg("Reloading L.I.S.A Engine")
        sys.path = self.syspath
        enabled_plugins = []
        self.build_activeplugins()

    def SchedReload(self):
        global taskman
        log.msg("Reloading Task Scheduler")
        self.taskman = taskman
        return self.taskman.reload()

LisaInstance = LisaFactory()
LisaProtocolInstance = Lisa(LisaInstance, LisaInstance.wit)

def Initialize():
    # Create the default core_intents_list intent
    defaults_intent_list = {'name': "core_intents_list",
                     'function': "list",
                     'module': "core.intents.Intents",
                     'enabled': True
    }

    intent_list, created = Intent.objects.get_or_create(name='core_intents_list', defaults=defaults_intent_list)

    # Create the default rule of the rule engine
    defaults_rule = {'name': "DefaultAnwser",
                     'order': 999,
                     'before': None,
                     'after': """lisaprotocol.answerToClient(json.dumps(
                                                {
                                                    'plugin': jsonOutput['plugin'],
                                                    'method': jsonOutput['method'],
                                                    'body': jsonOutput['body'],
                                                    'clients_zone': ['sender'],
                                                    'from': jsonOutput['from']
                                                }))""",
                     'end': True,
                     'enabled': True
    }
    default_rule, created = Rule.objects.get_or_create(name='DefaultAnwser', defaults=defaults_rule)


import requests

class Wit():
    def __init__(self, configuration):
        self.configuration = configuration
        self.failed_response = {"fr": "Je n'ai pas compris votre demande",
                   "en": "I didn't understood your question"
        }

    def message_send(self, text):
        headers = {'Authorization': 'Bearer ' + self.configuration['wit_token']}
        r = requests.get(self.configuration['wit_url'] + 'message', params={'q': text}, headers=headers)
        if r.ok:
            return r.json()

    def list_intents(self):
        headers = {'Authorization': 'Bearer ' + self.configuration['wit_token']}
        r = requests.get(self.configuration['wit_url'] + 'intents', headers=headers)
        if r.ok:
            return r.json()


from django.conf.urls import patterns, include, url
from tastypie.api import Api
from lisa.server.web.manageplugins.api import PluginResource, IntentResource
from lisa.server.web.interface.api import WidgetResource, WorkspaceResource, WidgetByUserResource
from api import LisaResource, UserResource
from twisted.python.reflect import namedAny
import tastypie_swagger

v1_api = Api(api_name='v1')
v1_api.register(UserResource())
v1_api.register(PluginResource())
v1_api.register(IntentResource())
v1_api.register(WorkspaceResource())
v1_api.register(WidgetResource())
v1_api.register(WidgetByUserResource())
v1_api.register(LisaResource())

urlpatterns = patterns('',
    url(r'^speech/', include('lisa.server.web.googlespeech.urls')),
    url(r'^plugins/', include('lisa.server.web.manageplugins.urls')),
    url(r'', include('lisa.server.web.interface.urls')),
)

from lisa.server.libs.server import enabled_plugins
for plugin in enabled_plugins:
    #try:
    urlpatterns += patterns('', url(r'^'+ str(plugin.lower()) + r'/', include(str(plugin) + '.web.urls')))
    v1_api.register(namedAny(plugin + '.web.api.' + plugin + 'Resource')())
    #except:
    #    pass

#Register plugin's API
urlpatterns += patterns('',
    url(r'^api/', include(v1_api.urls)),
    url(r'^api/docs/', include('tastypie_swagger.urls', namespace='tastypie_swagger')),
)
